import pkg_resources
import importlib
from requests import Session
from bs4 import BeautifulSoup

from jadi import component
from aj.plugins.packages.api import PackageManager, Package


@component(PackageManager)
class PIPPackageManager(PackageManager):
    """
    Manager to handle pip packages.
    """

    id = 'pip'
    name = 'PIP'

    def __init__(self, context):
        PackageManager.__init__(self, context)

    def __make_package_pipdist(self, dist):
        """
        Convert pip package object in package object.

        :param apt_package: Pip package object from apt module
        :type apt_package: Pip package object from apt module
        :return: Package object
        :rtype: Package object
        """

        p = Package(self)
        p.id = '%s==%s' % (dist.key, dist.version)
        p.name = dist.key
        p.version = dist.version
        p.description = None
        p.is_installed = True
        p.installed_version = dist.version
        return p

    def __make_package(self, dist):
        """
        Convert pip package object in package object.

        :param apt_package: Pip package object from apt module
        :type apt_package: Pip package object from apt module
        :return: Package object
        :rtype: Package object
        """

        p = Package(self)
        p.id = '%s==%s' % (dist['name'], dist['version'])
        p.name = dist['name']
        p.version = dist['version']
        p.description = dist['summary']
        p.released = dist['released']
        importlib.reload(pkg_resources)
        for d in pkg_resources.working_set:
            if d.key == p.name:
                p.is_installed = True
                p.installed_version = d.version
                p.is_upgradeable = p.installed_version != p.version
        return p

    def __requests_pypi_search(self, query):
        """
        Launch a query on pypi search and filter the results with bs4.
        Heavily inspired from https://github.com/victorgarric/pip_search/blob/master/pip_search/pip_search.py

        :param query: string to search for
        :type query: string
        :return:
        :rtype:
        """

        api_url = 'https://pypi.org/search/'
        session = Session()

        packages = []
        snippets = []
        for page in range(1,3):
            params = {'q': query, 'page': page}
            result = session.get(api_url, params=params)

            soup = BeautifulSoup(result.text, 'html.parser')
            snippets += soup.select('a[class*="snippet"]')

        for snippet in snippets:
            package = {}
            for value in ["name", "version", "released"]:
                package[value] = snippet.select_one(f'span[class*="{value}"]').text.strip()
            package['summary'] = snippet.select_one(f'p[class*="description"]').text.strip()
            packages.append(package)

        return packages

    def list(self, query=None):
        """
        Generator for all packages.

        :param query: Search string
        :type query: string
        :return: Package object
        :rtype:Package object
        """

        for dist in self.__requests_pypi_search(query):
            yield self.__make_package(dist)

    def get_package(self, _id):
        """
        Get package informations.

        :param _id: Package name
        :type _id: string
        :return: Package object
        :rtype: Package object
        """

        for d in pkg_resources.working_set:
            if d.key == _id.split('==')[0]:
                return self.__make_package_pipdist(d)

    def update_lists(self, progress_callback):
        """
        Refresh list of packages.

        :param progress_callback: Callback function to follow progress
        :type progress_callback: function
        """

        pass

    def get_apply_cmd(self, selection):
        """
        Prepare command to apply.

        :param selection: Dict of packages an actions
        :type selection: dict
        :return: Command for terminal use
        :rtype: string
        """

        packages = []
        cmd = ''

        for sel in selection:
            if sel['operation'] in ['install', 'upgrade']:
                packages.append(sel['package']['id'])
        if packages:
            cmd = 'python3 -m pip install ' + ' '.join(packages) + ' ;'
            packages = []

        for sel in selection:
            if sel['operation'] in ['remove']:
                packages.append(sel['package']['name'])
        if packages:
            cmd = 'python3 -m pip uninstall ' + ' '.join(packages)
        return cmd



# pyflakes: disable-all
from .api import *
from .main import *
from .views import *
from .tasks import *

try:
    from .managers.apt_manager import *
except ImportError:
    pass

try:
    from .managers.yum_manager import *
except ImportError:
    pass

from .managers.pip_manager import *



import yum

from jadi import component
from aj.plugins.packages.api import PackageManager, Package


@component(PackageManager)
class YUMPackageManager(PackageManager):
    """
    Manager to handle rpm packages.
    """

    id = 'yum'
    name = 'YUM'

    @classmethod
    def __verify__(cls):
        try:
            yum.YumBase().doGenericSetup(cache=1)
            return True
        except Exception as e:
            return False

    def __init__(self, context):
        PackageManager.__init__(self, context)
        self.yum = yum.YumBase()
        self.yum.doGenericSetup(cache=1)

    def __make_package(self, pkg):
        """
        Convert yum package object in package object.

        :param apt_package: Yum package object from apt module
        :type apt_package: Yum package object from apt module
        :return: Package object
        :rtype: Package object
        """

        pkg_installed = (self.yum.rpmdb.searchNames(names=[pkg.name]) or [None])[0]
        p = Package(self)
        p.id = '%s.%s' % (pkg.name, pkg.arch)
        p.name = pkg.name
        p.version = pkg.version
        p.description = pkg.arch  # nothing better
        p.is_installed = pkg_installed is not None
        if p.is_installed:
            p.installed_version = pkg_installed.version
            p.is_upgradeable = p.installed_version != p.version
        return p

    def list(self, query=None):
        """
        Generator for all packages.

        :param query: Search string
        :type query: string
        :return: Package object
        :rtype:Package object
        """

        for pkg in self.yum.pkgSack.returnPackages():
            yield self.__make_package(pkg)

    def get_package(self, _id):
        """
        Get package informations.

        :param _id: Package name
        :type _id: string
        :return: Package object
        :rtype: Package object
        """

        pkg = (self.yum.searchNames(names=[_id]) or [None])[0]
        return self.__make_package(pkg)

    def update_lists(self, progress_callback):
        """
        Refresh list of packages.

        :param progress_callback: Callback function to follow progress
        :type progress_callback: function
        """

        class Progress():
            def __init__(self):
                self.size = 0
                self.done = 0
                self.name = None

            def end(self, amount_read, now=None):
                pass

            def start(self, filename=None, url=None, basename=None, size=None, now=None, text=None):
                self.size = size
                self.done = 0
                self.name = url

            def update(self, amount_read, now=None):
                self.done = amount_read
                message = '%s%% %s' % (int(100 * self.done / self.size), self.name)
                progress_callback(message=message, done=self.done, total=self.size)

        progress_callback(message='Preparing')
        y = yum.YumBase()
        y.repos.setProgressBar(Progress())
        y.cleanMetadata()
        y.repos.doSetup()
        y.repos.populateSack()

    def get_apply_cmd(self, selection):
        """
        Prepare command to apply.

        :param selection: Dict of packages an actions
        :type selection: dict
        :return: Command for terminal use
        :rtype: string
        """

        to_install = [
            sel['package']['id']
            for sel in selection
            if sel['operation'] in ['install', 'upgrade']
        ]
        to_remove = [
            sel['package']['id']
            for sel in selection
            if sel['operation'] == 'remove'
        ]
        cmd = ''
        if len(to_install) > 0:
            cmd += 'yum install ' + ' '.join(to_install)
            if len(to_remove) > 0:
                cmd += ' && '
        if len(to_remove) > 0:
            cmd += 'yum remove ' + ' '.join(to_remove)
        return cmd



"""
Module to list, install and remove package on the current system.
"""

import json

from jadi import component
from aj.api.http import url, HttpPlugin
from aj.auth import authorize
from aj.api.endpoint import endpoint
from aj.plugins.packages.api import PackageManager


@component(HttpPlugin)
class Handler(HttpPlugin):
    def __init__(self, context):
        self.context = context
        self.managers = {x.id:x for x in PackageManager.all(self.context, ignore_exceptions=True)}

    def __package_to_json(self, package):
        """
        Convert package object into dict for convenience in angular.

        :param package: Package object
        :type package: Package object
        :return: Dict of informations from the package
        :rtype: dict
        """

        return {
            'id': package.id,
            'name': package.name,
            'version': package.version,
            'description': package.description,
            'released': package.released,
            'managerId': package.manager.id,
            'isInstalled': package.is_installed,
            'isUpgradeable': package.is_upgradeable,
            'installedVersion': package.installed_version,
        }

    @url(r'/api/packages/managers')
    @endpoint(api=True)
    def handle_api_managers(self, http_context):
        """
        List all availables managers.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :return: List of manager, one per dict
        :rtype: list of dict
        """

        return [
            {
                'id': mgr.id,
                'name': mgr.name,
            } for mgr in self.managers.values()
        ]

    @url(r'/api/packages/list/(?P<manager_id>\w+)')
    @endpoint(api=True)
    def handle_api_list(self, http_context, manager_id=None):
        """
        Filter installed packages names with query string.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param manager_id: Name of the manager, e.g. apt
        :type manager_id: string
        :return: List of packages informations, one per dict
        :rtype: list of dict
        """

        query = http_context.query['query']
        if len(query) < 3:
            raise Exception('Query too short')
        query = query.lower().strip()
        return [
            self.__package_to_json(package)
            for package in self.managers[manager_id].list(query=query)
            if query in package.id or query in package.name
        ]

    @url(r'/api/packages/get/(?P<manager_id>\w+)/(?P<package_id>\w+)')
    @endpoint(api=True)
    def handle_api_get(self, http_context, manager_id=None, package_id=None):
        """
        Get informations for one specific package, in one specific manager.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param manager_id: Name of the manager, e.g. apt
        :type manager_id: string
        :param package_id: Name of the package, e.g. cups
        :type package_id: string
        :return: Package informations
        :rtype: dict
        """

        return self.__package_to_json(self.managers[manager_id].get_package(package_id))

    @url(r'/api/packages/apply/(?P<manager_id>\w+)')
    @authorize('packages:install')
    @endpoint(api=True)
    def handle_api_apply(self, http_context, manager_id=None):
        """
        Prepare command to apply in order to launch it i na terminal.
        Commands may be install, upgrade, remove.
        Packages and commands are given in post body.
        Method POST.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param manager_id: Name of the manager, e.g. apt
        :type manager_id: strign
        :return: Terminal command
        :rtype: dict
        """

        mgr = self.managers[manager_id]
        selection = json.loads(http_context.body.decode())
        cmd = mgr.get_apply_cmd(selection)
        return {
            'terminalCommand': cmd,
        }
